==> lsp_types-0.5.0-py3-none-any.whl: lsp_types/session.py <==
import asyncio
import dataclasses as dc
import itertools
import json
import logging
import os
from typing import Any, Mapping

import lsp_types

StringDict = dict[str, Any]
PayloadLike = list[StringDict] | StringDict | Mapping[str, Any] | None
CONTENT_LENGTH = "Content-Length: "
ENCODING = "utf-8"


logger = logging.getLogger("lsp-types")


@dc.dataclass(kw_only=True)
class ProcessLaunchInfo:
    cmd: list[str]
    env: dict[str, str] = dc.field(default_factory=dict)
    cwd: str = os.getcwd()


class Error(Exception):
    def __init__(self, code: lsp_types.ErrorCodes, message: str) -> None:
        super().__init__(message)
        self.code = code

    def to_lsp(self) -> StringDict:
        return {"code": self.code, "message": super().__str__()}

    @classmethod
    def from_lsp(cls, d: StringDict) -> "Error":
        return Error(d["code"], d["message"])

    def __str__(self) -> str:
        return f"{super().__str__()} ({self.code})"


class LSPSession:
    """
    A session manager for Language Server Protocol communication.
    Provides async/await interface for requests and notification queue for handling server messages.

    Usage:
        async with LSPSession(process_info) as session:
            # Send request and await response
            init_result = await session.send.initialize(params)

            # Send notifications (awaiting is optional)
            await session.send.did_open_text_document(params)
            session.notify.did_change_text_document(params)

            # Process notifications from server
            async for notification in session.notifications():
                method = notification["method"]
                params = notification["params"]
                # Handle notification
    """

    def __init__(self, process_launch_info: ProcessLaunchInfo):
        self.process_launch_info = process_launch_info
        self.process: asyncio.subprocess.Process | None = None
        self._notification_queue: asyncio.Queue[StringDict] = asyncio.Queue()
        self._pending_requests: dict[int | str, asyncio.Future[Any]] = {}
        self._request_id_gen = itertools.count(1)
        self._tasks: list[asyncio.Task] = []
        self._shutdown = False

        # Maintain typed interface
        self.send = lsp_types.Request(self._send_request)
        self.notify = lsp_types.Notification(self._send_notification)

    async def __aenter__(self) -> "LSPSession":
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.stop()

    async def start(self) -> None:
        """Start the LSP server process and initialize communication."""
        child_proc_env = os.environ.copy()
        child_proc_env.update(self.process_launch_info.env)

        self.process = await asyncio.create_subprocess_exec(
            *self.process_launch_info.cmd,
            stdout=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=child_proc_env,
            cwd=self.process_launch_info.cwd,
        )

        self._tasks.extend(
            [
                asyncio.create_task(self._read_stdout()),
                asyncio.create_task(self._read_stderr()),
            ]
        )

    async def stop(self) -> None:
        """Stop the LSP server and clean up resources."""
        if not self._shutdown:
            try:
                await self.send.shutdown()
                await self.notify.exit()
            except ConnectionResetError:
                pass  # Server already closed

            self._shutdown = True

        for task in self._tasks:
            task.cancel()

        if self.process:
            try:
                return_code = await asyncio.wait_for(self.process.wait(), timeout=5.0)
                if return_code != 0:
                    logging.warning("Server exited with return code: %d", return_code)
            except asyncio.TimeoutError:
                try:
                    self.process.kill()
                except ProcessLookupError:
                    pass
            self.process = None

    async def notifications(self):
        """
        An async generator for processing server notifications.

        Usage:
            async for notification in session.notifications():
                # Process notification
        """
        while True:
            yield await self._notification_queue.get()
            self._notification_queue.task_done()

    async def _send_request(self, method: str, params: Mapping | None = None) -> Any:
        """Send a request to the server and await the response."""
        if not self.process or not self.process.stdin:
            raise RuntimeError("LSP process not available")

        request_id = next(self._request_id_gen)

        future: asyncio.Future[Any] = asyncio.Future()
        self._pending_requests[request_id] = future

        payload = make_request(method, request_id, params)
        await self._send_payload(self.process.stdin, payload)

        try:
            return await future
        finally:
            self._pending_requests.pop(request_id, None)

    def _send_notification(
        self, method: str, params: Mapping | None = None
    ) -> asyncio.Task[None]:
        """Send a notification to the server."""
        if not self.process or not self.process.stdin:
            logging.warning("LSP process not available: [%s]", method)
            return asyncio.create_task(asyncio.sleep(0))

        payload = make_notification(method, params)
        task = asyncio.create_task(self._send_payload(self.process.stdin, payload))
        self._tasks.append(task)

        return task

    @staticmethod
    async def _send_payload(stream: asyncio.StreamWriter, payload: StringDict) -> None:
        """Send a payload to the server asynchronously."""
        logger.debug("Client -> Server: %s", payload)

        body = json.dumps(
            payload, check_circular=False, ensure_ascii=False, separators=(",", ":")
        ).encode(ENCODING)
        message = (
            f"Content-Length: {len(body)}\r\n",
            "Content-Type: application/vscode-jsonrpc; charset=utf-8\r\n\r\n",
        )

        # TODO: Maybe use a lock to avoid interleaving messages?
        stream.writelines([part.encode(ENCODING) for part in message] + [body])
        await stream.drain()

    async def _read_stdout(self) -> None:
        """Read and process messages from the server's stdout."""
        try:
            while (
                self.process
                and self.process.stdout
                and not self.process.stdout.at_eof()
            ):
                # Read header
                line = await self.process.stdout.readline()
                if not line.strip():
                    continue

                content_length = 0
                if line.startswith(b"Content-Length: "):
                    content_length = int(line.split(b":")[1].strip())

                if not content_length:
                    continue

                while line and line.strip():
                    line = await self.process.stdout.readline()

                # Read message body
                body = await self.process.stdout.readexactly(content_length)
                payload = json.loads(body.strip())

                logger.debug("Server -> Client: %s", payload)

                # Handle message based on type
                if "method" in payload:
                    # Server notification
                    await self._notification_queue.put(payload)
                elif "id" in payload:
                    # Response to client request
                    request_id = payload["id"]
                    future = self._pending_requests.get(request_id)
                    if future:
                        if "result" in payload:
                            future.set_result(payload["result"])
                        elif "error" in payload:
                            future.set_exception(Error.from_lsp(payload["error"]))
                        else:
                            future.set_exception(
                                Error(
                                    lsp_types.ErrorCodes.InvalidRequest,
                                    "Invalid response",
                                )
                            )
        except asyncio.CancelledError:
            return
        except Exception:
            logger.exception("Client - Error reading stdout")

    async def _read_stderr(self) -> None:
        """Read and log messages from the server's stderr."""
        try:
            while (
                self.process
                and self.process.stderr
                and not self.process.stderr.at_eof()
            ):
                line = await self.process.stderr.readline()
                if not line:
                    continue
                logger.error(f"Server - stderr: {line.decode(ENCODING).strip()}")
        except asyncio.CancelledError:
            return
        except Exception:
            logger.exception("Client - Error reading stderr")


def make_notification(method: str, params: PayloadLike) -> StringDict:
    return {"jsonrpc": "2.0", "method": method, "params": params}


def make_request(method: str, request_id: int | str, params: PayloadLike) -> StringDict:
    return {"jsonrpc": "2.0", "method": method, "id": request_id, "params": params}

==> lsp_types-0.6.0-py3-none-any.whl: lsp_types/session.py <==
import asyncio
import dataclasses as dc
import itertools
import json
import logging
import os
from typing import Any, cast as type_cast

import lsp_types

CONTENT_LENGTH = "Content-Length: "
ENCODING = "utf-8"


logger = logging.getLogger("lsp-types")


@dc.dataclass(kw_only=True)
class ProcessLaunchInfo:
    cmd: list[str]
    env: dict[str, str] = dc.field(default_factory=dict)
    cwd: str = os.getcwd()


class Error(Exception):
    def __init__(self, code: lsp_types.ErrorCodes, message: str) -> None:
        super().__init__(message)
        self.code = code

    def to_lsp(self) -> lsp_types.LSPObject:
        return {"code": self.code, "message": super().__str__()}

    @classmethod
    def from_lsp(cls, d: lsp_types.LSPObject) -> "Error":
        code = lsp_types.ErrorCodes(d["code"])
        message = type_cast(str, d["message"])
        return Error(code, message)

    def __str__(self) -> str:
        return f"{super().__str__()} ({self.code})"


class LSPSession:
    """
    A session manager for Language Server Protocol communication.
    Provides async/await interface for requests and notification queue for handling server messages.

    Usage:
        async with LSPSession(process_info) as session:
            # Send request and await response
            init_result = await session.send.initialize(params)

            # Send notifications (awaiting is optional)
            await session.send.did_open_text_document(params)
            session.notify.did_change_text_document(params)

            # Process notifications from server
            async for notification in session.notifications():
                method = notification["method"]
                params = notification["params"]
                # Handle notification
    """

    def __init__(self, process_launch_info: ProcessLaunchInfo):
        self._process_launch_info = process_launch_info
        self._process: asyncio.subprocess.Process | None = None
        self._notification_listeners: list[asyncio.Queue[lsp_types.LSPObject]] = []
        self._pending_requests: dict[int | str, asyncio.Future[Any]] = {}
        self._request_id_gen = itertools.count(1)
        self._tasks: list[asyncio.Task] = []
        self._shutdown = False

        # Maintain typed interface
        self.send = lsp_types.RequestFunctions(self._send_request)
        self.notify = lsp_types.NotificationFunctions(
            self._send_notification, self._on_notification
        )

    async def __aenter__(self) -> "LSPSession":
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.stop()

    async def start(self) -> None:
        """Start the LSP server process and initialize communication."""
        if self._process:
            raise RuntimeError("LSP process already started")

        child_proc_env = os.environ.copy()
        child_proc_env.update(self._process_launch_info.env)

        self._process = await asyncio.create_subprocess_exec(
            *self._process_launch_info.cmd,
            stdout=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=child_proc_env,
            cwd=self._process_launch_info.cwd,
        )

        self._tasks.extend(
            [
                asyncio.create_task(self._read_stdout()),
                asyncio.create_task(self._read_stderr()),
            ]
        )

    async def stop(self) -> None:
        """Stop the LSP server and clean up resources."""
        if not self._shutdown:
            try:
                await self.send.shutdown()
                await self.notify.exit()
            except ConnectionResetError:
                pass  # Server already closed

            self._shutdown = True

        for task in self._tasks:
            task.cancel()

        if self._process:
            try:
                return_code = await asyncio.wait_for(self._process.wait(), timeout=5.0)
                if return_code != 0:
                    logger.warning("Server exited with return code: %d", return_code)
            except asyncio.TimeoutError:
                try:
                    self._process.kill()
                except ProcessLookupError:
                    pass
            self._process = None

    async def notifications(self):
        """
        An async generator for processing server notifications.

        Usage:
            async for notification in session.notifications():
                # Process notification
        """
        queue: asyncio.Queue[lsp_types.LSPObject] = asyncio.Queue()
        self._notification_listeners.append(queue)

        try:
            while True:
                yield await queue.get()
                queue.task_done()
        finally:
            self._notification_listeners.remove(queue)

    async def _send_request(self, method: str, params: lsp_types.LSPAny = None) -> Any:
        """Send a request to the server and await the response."""
        if not self._process or not self._process.stdin:
            raise RuntimeError("LSP process not available")

        request_id = next(self._request_id_gen)

        future: asyncio.Future[Any] = asyncio.Future()
        self._pending_requests[request_id] = future

        payload = make_request(method, request_id, params)
        await self._send_payload(self._process.stdin, payload)

        try:
            return await future
        finally:
            self._pending_requests.pop(request_id, None)

    def _send_notification(
        self, method: str, params: lsp_types.LSPAny = None
    ) -> asyncio.Task[None]:
        """Send a notification to the server."""
        if not self._process or not self._process.stdin:
            logger.warning("LSP process not available: [%s]", method)
            return asyncio.create_task(asyncio.sleep(0))

        payload = make_notification(method, params)
        task = asyncio.create_task(self._send_payload(self._process.stdin, payload))
        self._tasks.append(task)

        return task

    def _on_notification(
        self, method: str, timeout: float | None = None
    ) -> asyncio.Future[lsp_types.LSPAny]:
        """Wait for a specific notification from the server."""

        async def _wait_for_notification():
            async for notification in self.notifications():
                if notification["method"] == method:
                    return notification["params"]

        coroutine = _wait_for_notification()
        if timeout is not None:
            coroutine = asyncio.wait_for(coroutine, timeout)

        wait_task = asyncio.create_task(coroutine)
        self._tasks.append(wait_task)
        wait_task.add_done_callback(self._tasks.remove)

        return wait_task

    @staticmethod
    async def _send_payload(
        stream: asyncio.StreamWriter, payload: lsp_types.LSPObject
    ) -> None:
        """Send a payload to the server asynchronously."""
        logger.debug("Client -> Server: %s", payload)

        body = json.dumps(
            payload, check_circular=False, ensure_ascii=False, separators=(",", ":")
        ).encode(ENCODING)
        message = (
            f"Content-Length: {len(body)}\r\n",
            "Content-Type: application/vscode-jsonrpc; charset=utf-8\r\n\r\n",
        )

        # TODO: Maybe use a lock to avoid interleaving messages?
        stream.writelines([part.encode(ENCODING) for part in message] + [body])
        await stream.drain()

    async def _read_stdout(self) -> None:
        """Read and process messages from the server's stdout."""
        try:
            while (
                self._process
                and self._process.stdout
                and not self._process.stdout.at_eof()
            ):
                # Read header
                line = await self._process.stdout.readline()
                if not line.strip():
                    continue

                content_length = 0
                if line.startswith(b"Content-Length: "):
                    content_length = int(line.split(b":")[1].strip())

                if not content_length:
                    continue

                while line and line.strip():
                    line = await self._process.stdout.readline()

                # Read message body
                body = await self._process.stdout.readexactly(content_length)
                payload = json.loads(body.strip())

                logger.debug("Server -> Client: %s", payload)

                # Handle message based on type
                if "method" in payload:
                    # Server notification
                    [q.put_nowait(payload) for q in self._notification_listeners]
                elif "id" in payload:
                    # Response to client request
                    request_id = payload["id"]
                    future = self._pending_requests.get(request_id)
                    if future:
                        if "result" in payload:
                            future.set_result(payload["result"])
                        elif "error" in payload:
                            future.set_exception(Error.from_lsp(payload["error"]))
                        else:
                            future.set_exception(
                                Error(
                                    lsp_types.ErrorCodes.InvalidRequest,
                                    "Invalid response",
                                )
                            )
        except asyncio.CancelledError:
            return
        except Exception:
            logger.exception("Client - Error reading stdout")

    async def _read_stderr(self) -> None:
        """Read and log messages from the server's stderr."""
        try:
            while (
                self._process
                and self._process.stderr
                and not self._process.stderr.at_eof()
            ):
                line = await self._process.stderr.readline()
                if not line:
                    continue
                logger.error(f"Server - stderr: {line.decode(ENCODING).strip()}")
        except asyncio.CancelledError:
            return
        except Exception:
            logger.exception("Client - Error reading stderr")


def make_notification(method: str, params: lsp_types.LSPAny) -> lsp_types.LSPObject:
    return {"jsonrpc": "2.0", "method": method, "params": params}


def make_request(
    method: str, request_id: int | str, params: lsp_types.LSPAny
) -> lsp_types.LSPObject:
    return {"jsonrpc": "2.0", "method": method, "id": request_id, "params": params}
